/*
 * The "official" fmc-tdc API
 *
 * Copyright (C) 2012-2018 CERN (www.cern.ch)
 * Author: Tomasz Włostowski <tomasz.wlostowski@cern.ch>
 * Author: Alessandro Rubini <rubini@gnudd.com>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */
#ifndef __FMCTDC_LIB_H__
#define __FMCTDC_LIB_H__

#ifdef __cplusplus
#pragma GCC diagnostic ignored "-Wwrite-strings"
extern "C" {
#endif

#include <inttypes.h>
#include <stdint.h>

/**
 * printf format for timestamps with pico-second resolution
 */
#define PRItsps "%010"PRIu64"s  %012"PRIu64"ps"

/**
 * printf value for timestamps with pico-second resolution
 */
#define PRItspsVAL(_ts) (_ts)->seconds, (uint64_t)(((uint64_t)(_ts)->coarse * 8000ULL) + ((uint64_t)(_ts)->frac *8000ULL / 4096ULL))

/**
 * printf format for timestamps with White-Rabbit notation
 */
#define PRItswr "%10"PRIu64":%09u:%04u"

/**
 * printf value for timestamp with White-Rabbit notation
 */
#define PRItswrVAL(_ts) (_ts)->seconds, (_ts)->coarse, (_ts)->frac


#define __FMCTDC_ERR_MIN 4096
enum fmctdc_error_numbers {
	FMCTDC_ERR_VMALLOC = __FMCTDC_ERR_MIN,
	FMCTDC_ERR_UNKNOWN_BUFFER_TYPE,
	FMCTDC_ERR_NOT_CONSISTENT_BUFFER_TYPE,
	FMCTDC_ERR_VERSION_MISMATCH,
	__FMCTDC_ERR_MAX,
};

/**
 * Enumeration for all TDC channels
 */
enum fmctdc_channel {
	FMCTDC_CH_1 = 0,
	FMCTDC_CH_2,
	FMCTDC_CH_3,
	FMCTDC_CH_4,
	FMCTDC_CH_5,
	FMCTDC_CH_LAST = FMCTDC_CH_5,
	FMCTDC_NUM_CHANNELS = 5
};

/**
 * Enumeration of all buffer modes
 */
enum fmctdc_buffer_sw_mode {
	FMCTDC_BUF_SW_MODE_FIFO = 0, /**< FIFO policy: when buffer is full, new
					time-stamps will be dropped */
	FMCTDC_BUF_SW_MODE_CIRC, /**< circular buffer policy: when the buffer is
				    full, old time-stamps will be overwritten by
				    new ones */
};

/**
 * Enumeration of all buffer types
 */
enum fmctdc_buffer_type {
	FMCTDC_BUFFER_KMALLOC = 0, /**< kernel allocator: kmalloc */
	FMCTDC_BUFFER_VMALLOC, /**< kernel allocator: vmalloc */
};

/**
 * Enumeration for all possible status of a channel
 */
enum fmctdc_channel_status {
	FMCTDC_STATUS_DISABLE = 0, /**< The cannel is disable */
	FMCTDC_STATUS_ENABLE, /**< the channel is enable */
};

/**
 * Enumeration of all hardware buffer mode
 */
enum fmctdc_buffer_hw_mode {
	FMCTDC_BUF_HW_MODE_FIFO = 0,
	FMCTDC_BUF_HW_MODE_DMA,
};

/**
 * Enumeration for all possible time-stmap processing mode
 */
enum fmctdc_process_mode {
	FMCTDC_TS_MODE_RAW = 0, /**< directly from ACAM chip. It does not do
				   any time-stamp processing */
	FMCTDC_TS_MODE_OFFSETS, /**< it applies all offsets (calibration and
				   user) */
};

/**
 * Enumeration of all possible White-Rabbit status
 */
enum fmctdc_wr_status {
	FMCTDC_WR_OFF = 0, /**< White-Rabbit disabled */
	FMCTDC_WR_ON_NOLINK, /**< Fiber connection missing */
	FMCTDC_WR_ON_SYNCING, /**< White-Rabbit is synchronizing */
	FMCTDC_WR_ON, /**< White-Rabbit is working */
};

/**
 * Opaque data type used as token. Do not try to access.
 */
struct fmctdc_board;


/**
 * FMC-TDC time-stamp descriptor
 */
struct fmctdc_time {
	uint64_t seconds; /**< TAI seconds. Note this is *not* an UTC time;
			     the counter does not support leap seconds. The
			     internal counter is also limited to 32 bits
			     (2038-error-prone). */
	uint32_t coarse; /**< number of ticks of 8ns since the beginning of
			    the last second*/
	uint32_t frac; /**< fractional part of an 8 ns tick, rescaled
			  to (0..4095) range - i.e. 0 = 0 ns, and
			  4095 = 7.999 ns. */
	uint32_t seq_id; /**< channel sequence number*/
	uint32_t debug;  /**< debug stuff, driver/firmware-specific */
};


/**
 * @defgroup sw_lib_init Library initialization and utilities
 * @{
 */
extern int fmctdc_init(void);
extern void fmctdc_exit(void);
extern const char *fmctdc_strerror(int err);
extern int fmctdc_temperature_get(struct fmctdc_board *userb,
				  float *temperature);
/**@}*/

/**
 * @defgroup sw_lib_acc Device Access
 * @{
 */
extern struct fmctdc_board *fmctdc_open(int dev_id);
extern struct fmctdc_board *fmctdc_open_by_lun(int lun);
extern int fmctdc_fileno_channel(struct fmctdc_board *b, unsigned int channel);
extern int fmctdc_close(struct fmctdc_board *);
/**@}*/

/**
 * @defgroup sw_lib_hw Hardware Configuration and Information
 * @{
 */
extern int fmctdc_temperature_get(struct fmctdc_board *userb,
				  float *temperature);
extern int fmctdc_coalescing_timeout_set(struct fmctdc_board *userb,
					 unsigned int channel,
					 unsigned int timeout_ms);
extern int fmctdc_coalescing_timeout_get(struct fmctdc_board *userb,
					 unsigned int channel,
					 unsigned int *timeout_ms);
extern int fmctdc_termination_set(struct fmctdc_board *b,
				  unsigned int channel,
				  unsigned int enable);
extern int fmctdc_termination_get(struct fmctdc_board *b,
				  unsigned int channel,
				  unsigned int *enable);
extern int fmctdc_buffer_hw_mode_get(struct fmctdc_board *userb,
				     unsigned int channel,
				     enum fmctdc_buffer_hw_mode *mode);
extern int fmctdc_stats_recv_get(struct fmctdc_board *userb,
				 unsigned int channel,
				 uint32_t *val);
extern int fmctdc_stats_trans_get(struct fmctdc_board *userb,
				  unsigned int channel,
				  uint32_t *val);
/**@}*/

/**
 * @defgroup sw_lib_buf Buffer
 * @{
 */
extern int fmctdc_flush(struct fmctdc_board *userb, unsigned int channel);
extern int fmctdc_buffer_type_get(struct fmctdc_board *userb);
extern int fmctdc_buffer_type_set(struct fmctdc_board *userb,
				  enum fmctdc_buffer_type type);
extern int fmctdc_buffer_mode_get(struct fmctdc_board *userb,
				     unsigned int channel,
				     enum fmctdc_buffer_sw_mode *mode);
extern int fmctdc_buffer_mode_set(struct fmctdc_board *userb,
				     unsigned int channel,
				     enum fmctdc_buffer_sw_mode mode);
extern int fmctdc_buffer_len_get(struct fmctdc_board *userb,
				 unsigned int channel,
				 unsigned int *length);
extern int fmctdc_buffer_len_set(struct fmctdc_board *userb,
				 unsigned int channel,
				 unsigned int length);
/**@}*/

/**
 * @defgroup sw_lib_tim Time
 * @{
 */
extern int fmctdc_time_set(struct fmctdc_board *b, struct fmctdc_time *t);
extern int fmctdc_time_get(struct fmctdc_board *b, struct fmctdc_time *t);
extern int fmctdc_time_set_host(struct fmctdc_board *userb);
extern int fmctdc_wr_status_set(struct fmctdc_board *userb, int on);
extern int fmctdc_wr_status_get(struct fmctdc_board *userb,
				enum fmctdc_wr_status *status);
extern int fmctdc_process_set(struct fmctdc_board *userb,
			      unsigned int channel,
			      enum fmctdc_process_mode mode);
extern int fmctdc_process_get(struct fmctdc_board *userb,
			      unsigned int channel,
			      enum fmctdc_process_mode *mode);
extern int fmctdc_offset_set(struct fmctdc_board *userb,
				unsigned int channel, int32_t offset);
extern int fmctdc_offset_get(struct fmctdc_board *userb,
				unsigned int channel, int32_t *offset);
extern int fmctdc_fread(struct fmctdc_board *b, unsigned int channel,
			struct fmctdc_time *t, int n);
extern int fmctdc_read(struct fmctdc_board *b, unsigned int channel,
		       struct fmctdc_time *t, int n, int flags);
extern int fmctdc_status_set(struct fmctdc_board *userb,
			     unsigned int channel,
			     enum fmctdc_channel_status status);
extern int fmctdc_status_get(struct fmctdc_board *userb,
			      unsigned int channel,
			     enum fmctdc_channel_status *status);
extern int fmctdc_ts_sub(struct fmctdc_time *r,
			 const struct fmctdc_time *a,
			 const struct fmctdc_time *b);
extern void fmctdc_ts_add(struct fmctdc_time *r,
			  const struct fmctdc_time *a,
			  const struct fmctdc_time *b);
extern void fmctdc_ts_norm(struct fmctdc_time *a);
extern uint64_t fmctdc_ts_approx_ns(struct fmctdc_time *a);
extern uint64_t fmctdc_ts_ps(struct fmctdc_time *a);
/**@}*/

/* other */
extern const char * const libfmctdc_version_s;
extern const char * const libfmctdc_zio_version_s;

#ifdef __cplusplus
}
#endif

#endif /* __FMCTDC_LIB_H__ */
