/* Copyright 2018 CERN
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * License: GPLv2
 *
 * This is an example program that shows the different part of the library
 * in action.
 *
 *
 * This is part of the documentation, so when you change it **REMEMBER** to
 * update the references in documents in `doc/`.
 *
 * The reason to have the example here as a true program is that we can always
 * compile it and run it, so that we can validate the example.
 * This way, at least for this aspect, the documentation in always aligned.
 *
 * Do not try to optimize this code, the purpose is to show code, highlight
 * snippet and being able to refer to it from the documentation
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <poll.h>
#include <inttypes.h>

#include "fmctdc-lib.h"

static void __status(struct fmctdc_board *tdc,
		     unsigned int channel);
static int __acquire(struct fmctdc_board *tdc,
		     unsigned int channel);
static int __prepare_acquire(struct fmctdc_board *tdc,
			     unsigned int channel);
static int __config(struct fmctdc_board *tdc,
		    unsigned int channel);
static int __config_and_prepare_acquire(struct fmctdc_board *tdc,
					unsigned int channel);
static int __use_fmctdc_library(uint32_t dev_id,
				unsigned int channel);

int main(int argc, char *argv[])
{
	const unsigned int channel = FMCTDC_CH_1;
	int err, board_n;
	uint32_t dev_id;

	err = sscanf(argv[1], "0x%"SCNx32, &dev_id);
	if (err != 1)
		exit(EXIT_FAILURE);

	board_n = fmctdc_init();
	if (board_n < 0)
		exit(EXIT_FAILURE);
	err = __use_fmctdc_library(dev_id, channel);
	fmctdc_exit();

	exit(err ? EXIT_FAILURE : EXIT_SUCCESS);
}

static int __use_fmctdc_library(uint32_t dev_id, unsigned int channel)
{
	struct fmctdc_board *tdc;
	int err;

	tdc = fmctdc_open(dev_id);
	if (!tdc)
		return -1;
	err = __config_and_prepare_acquire(tdc, channel);
	fmctdc_close(tdc);

	if (err)
		fprintf(stderr, "Error: %s",
			fmctdc_strerror(errno));

	return err;
}

static int __config_and_prepare_acquire(struct fmctdc_board *tdc,
					unsigned int channel)
{
	int err;

	err = __config(tdc, channel);
	if (err)
		return err;

	__status(tdc, channel);

	err = __prepare_acquire(tdc, channel);
	if (err)
		return err;

	return 0;
}

static int __config(struct fmctdc_board *tdc, unsigned int channel)
{
	int err;

	/* hardware configuration */
	err = fmctdc_coalescing_timeout_set(tdc, channel, 10);
	if (err)
		return err;
	err = fmctdc_termination_set(tdc, channel, 1);
	if (err)
		return err;

	/* buffer configuration */
	err = fmctdc_buffer_mode_set(tdc, channel,
				     FMCTDC_BUF_SW_MODE_FIFO);
	if (err)
		return err;
	err = fmctdc_buffer_len_set(tdc, channel, 100);
	if (err)
		return err;

	/* timestamp processing */
	err = fmctdc_process_set(tdc, channel, FMCTDC_TS_MODE_OFFSETS);
	if (err)
		return err;
	err = fmctdc_offset_set(tdc, channel, 1000); /* +1ns */
	if (err)
		return err;

	/* timing configuration */
	err = fmctdc_wr_status_set(tdc, 1);
	if (err)
		return err;

	return 0;
}

static void __status(struct fmctdc_board *tdc, unsigned int channel)
{
	struct fmctdc_time time;
	unsigned int val;
	int32_t offset;
	int err;

	/* hardware configuration */
	err = fmctdc_coalescing_timeout_get(tdc, channel, &val);
	if (!err)
		fprintf(stdout, "Channel %u: coalescing timeout %u\n",
		        channel, val);

	err = fmctdc_termination_get(tdc, channel, &val);
	if (!err)
		fprintf(stdout, "Channel %u: termination %u\n",
		        channel, val);

	/* buffer configuration */
	err = fmctdc_buffer_mode_get(tdc, channel, &val);
	if (!err)
		fprintf(stdout, "Channel %u: buffer mode %u\n",
			channel, val);
	err = fmctdc_buffer_len_get(tdc, channel, &val);
	if (!err)
		fprintf(stdout, "Channel %u: buffer len %u\n",
			channel, val);


	/* timestamp processing */
	err = fmctdc_process_get(tdc, channel, &val);
	if (!err)
		fprintf(stdout, "Channel %u: processing mode %u\n",
			channel, val);
	err = fmctdc_offset_get(tdc, channel, &offset);
	if (!err)
		fprintf(stdout, "Channel %u: offset %u [ns]\n",
			channel, offset);

	/* timing configuration */
	err = fmctdc_wr_status_get(tdc, &val);
	if (!err)
		fprintf(stdout, "white Rabbit status: %u\n", val);
	err = fmctdc_time_get(tdc, &time);
	if (!err)
		fprintf(stdout, "Board time: "PRItsps" - "PRItswr"\n",
			PRItspsVAL(&time),
			PRItswrVAL(&time));
}


static int __prepare_acquire(struct fmctdc_board *tdc, unsigned int channel)
{
	int err;

	err = fmctdc_flush(tdc, channel);
	if (err)
		return err;

	err = fmctdc_status_set(tdc, channel, FMCTDC_STATUS_ENABLE);
	if (err)
		return err;

	err = __acquire(tdc, channel);

	err = fmctdc_status_set(tdc, channel, FMCTDC_STATUS_DISABLE);
	if (err)
		return err;

	return 0;
}

static int __acquire(struct fmctdc_board *tdc, unsigned int channel)
{
	struct fmctdc_time timestamp;
	struct pollfd p;
	int ret;

	p.fd = fmctdc_fileno_channel(tdc, channel);
	p.events = POLLIN | POLLERR;
	ret = poll(&p, 1, 10000);
	if (ret <= 0)
		return ret;

	ret = fmctdc_read(tdc, channel, &timestamp, 1, 0);
	if (ret < 0)
		return ret;
	if (ret == 0) {
		errno = EAGAIN;
		return -1;
	}

	fprintf(stdout, "Channel %d: "PRItsps" - "PRItswr"\n",
		channel,
		PRItspsVAL(&timestamp),
		PRItswrVAL(&timestamp));

	return 0;
}
